/****************************************************************************
 *
 *           Program : pi.c
 *
 *	     Author  : Kim, Dongmin
 *
 *
 *  This is a program that calculate pi by using Simpson's composite rule 
 *
 *  algorithm for integration in parallel.
 *
 *  By integrate the following formula :
 *
 *        / 1
 *	  |   ( 4 / ( 1 + ( x**2 ) ) ) dx
 *	  / 0
 *
 *  the pi can be calculated.
 *
 *  In this program, each processor calculates the sub area between 0 and 1,
 *
 *  and the processor 0 summerize them and display the final result.
 *
 *  According to the number of processors (n), each processor can have its
 *
 *  end points ( a & b).
 *
 *  With a, b, & M, each processor can calculate the value of
 *
 *        / a
 *	  |   ( 4 / ( 1 + ( x**2 ) ) ) dx
 *	  / b
 *
 *  The summation of these values are the pi.
 *
 *  For example, if we have 4 processors, then
 *
 *  0 processor will calculates between 0 and 0.25,
 *
 *  1 processor between 0.25 and 0.5, 2 processor between 0.5 and 0.75, and
 *
 *  3 processor between 0.75 and 1.
 *
 *  In this program, M ( basic calculation term ) is fixed to 100.
 *
 *  This means that each processor calculates the integration according to
 *
 *  ( a - b ) / ( 2 * M ), where a is the start point and b is the end point.
 *
 *  To complie : gcc pi.c -o pi -L/usr/local/lib/rs6000/ch_p4 -lmpi -lm
 *
 *  To run     : mpirun -mr_machinefile machinelist -mr_np 4 pi
 *
 ****************************************************************************/

#include	<stdio.h>
#include	<mpi.h>
#include        <math.h>

#define M       100
#define G_start 0.0  /* global start point */
#define G_end   1.0  /* global end point */

#define real_pi   ( (double) (4.0*atan(1.0)) ) /* the calculate value of PI */

/****************************************************************************
 *
 *         Procedure : f()
 *         Input     : double x
 *         Output    : double y
 *         Function  : calculate the y value using input x and the equation
 *                     4 / ( 1 + ( x**2 ) ).
 *                     e.g.
 *                     y = 4 / ( 1 + ( x**2 ) )
 *
 ****************************************************************************/
double  f( x )
     double x;
{
  double y;

  y = 4.0 / ( 1.0 + ( x * x ) );

  return y;
}

/****************************************************************************
 *
 *         Procedure : pi()
 *         Input     : double start_pt, double end_pt
 *         Output    : double summation
 *         Function  : find the summation between start_pt and end_pt
 *                     using Simpson's composite rule algorithm 
 *                     for integration.
 *
 *                     Algorithm :
 *
 *                     h = ( end_pt - start_pt ) / (2m)
 *                     XI0 = f(start_pt) + f(end_pt)
 *                     XI1 = 0; ( summation of f(x_2i-1) )
 *                     XI2 = 0; ( summation of f(x_2i) )
 *                     for i = 1 ; 2m -1 ; i++
 *                        X = start_pt I i * h
 *                        if i is even
 *                           XI2 = XI2 + f(X)
 *                        else
 *                           XI1 = XI1 + f(X)
 *                     XI = h * ( XI0 + 2 * XI2 + 4 * XI1 ) / 3
 *
 ****************************************************************************/
double pi(start_pt, end_pt)
     double start_pt, end_pt;
{
  double  h;
  double  XI0;  /* sum of f(start_pt) and f(end_pt) */
  double  XI1;  /* summation of f(x_2i-1) */
  double  XI2;  /* summation of f(x_2i) */
  double  XI;   /* the result */
  double  x;
  int     i;

  h = ( end_pt - start_pt ) / (double)( 2 * M );

  XI0 = f(start_pt) + f(end_pt);
  XI1 = 0;
  XI2 = 0;

  for ( i = 1 ; i < 2 * M ; i++ ) {
    x = start_pt + i * h;
    if ( ( i % 2 ) == 0 ) /* i is even */
      XI2 = XI2 + f(x);
    else
      XI1 = XI1 + f(x);
  }
  XI = h * ( XI0 + 2 * XI2 + 4 * XI1 ) / 3.0;

  return XI;
}

/****************************************************************************
 *
 *         Procedure : main()
 *         Function  : main routine of this program.
 *
 ****************************************************************************/
void main(argc, argv)
  int argc;
  char **argv;
{

   int     nprocs;      /* Number of processors */
   int     pnum;        /* Processor number starting from 0 */
   double  local_start; /* local start point */
   double  local_end;   /* local end point */
   double  term;        /* the length of term */
   double  sum;         /* summation between local_start and local_end */
   double  other;       /* result from other processor */
   double  error;       /* difference */
   int     from;
   MPI_Status  recv_status;

   MPI_Init(&argc, &argv);

   /* Who am I ? */
   MPI_Comm_rank(MPI_COMM_WORLD, &pnum);

   /* How many processors participate ? */
   MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

   /* calculate start point and end points */

   term = ( G_end - G_start ) / nprocs;

   local_start = G_start + pnum * term;
   local_end   = local_start + term;

   /* calculate the sum */
   sum = pi( local_start, local_end );

   if ( pnum == 0 )  { /* 0 processor collects the sums from other processors */

     for ( from = 1 ; from < nprocs ; from++ ) {

       /* 0 processor receives the sums from the other processors */
       MPI_Recv( &other, 1, MPI_DOUBLE, from, 99, MPI_COMM_WORLD, &recv_status);

       sum = sum + other;
     }
     error = real_pi - sum;
     printf("\n\n\n");
     printf("***************************************************************\n");
     printf("  PI using Simpson's composite rule algorithm for integration\n");
     printf("\n");
     printf("   Number of Processor(s) : %d\n", nprocs);
     printf("   PI ( by integration ) = %18.16f\n", sum);
     printf("   PI ( by calculation ) = %18.16f\n", real_pi);
     printf("   error = %18.16f\n", error);
     printf("***************************************************************\n");


   }
   else { /* Processors which is not 0 send the sum to the 0 processor */
     MPI_Send( &sum, 1, MPI_DOUBLE, 0, 99, MPI_COMM_WORLD);
   }

   MPI_Finalize();

}


